@@ -1,13 +1,13 @@
 from enum import Enum
 from fnmatch import fnmatch
-from typing import Optional, Container, FrozenSet, Iterable, List, Dict, Union, Any
+from itertools import chain
+from typing import Optional, Container, FrozenSet, Iterable, List, Dict, Any
 
 from pydantic import BaseModel
 
 __all__ = [
     "Feature",
     "FeatureProperties",
-    "FeatureTask",
     "ManifestInfo",
     "Modpack",
     "ManifestIndex",
@@ -17,13 +17,16 @@ __all__ = [
     "SimplePatternList",
     "SourceManifest",
     "TargetManifest",
-    "Task",
-    "UserFileTask",
-    "TargetTask",
+    "FileTask",
 ]
 
 
-class BaseManifest(BaseModel):
+class HashEqModel(BaseModel):
+    def __eq__(self, other):
+        return hash(self) == hash(other)
+
+
+class BaseManifest(HashEqModel):
     name: str
     title: str
     version: str = "1.0"
@@ -34,9 +37,6 @@ class BaseManifest(BaseModel):
     def __hash__(self):
         return hash((self.name, self.title, self.version))
 
-    def __eq__(self, other):
-        return hash(self) == hash(other)
-
 
 class ManifestInfo(BaseManifest):
     location: str
@@ -47,11 +47,8 @@ class ManifestInfo(BaseManifest):
     def __hash__(self):
         return hash((self.name, self.title, self.version, self.location))
 
-    def __eq__(self, other):
-        return hash(self) == hash(other)
 
-
-class ManifestIndex(BaseModel):
+class ManifestIndex(HashEqModel):
     minimumVersion: int = 3
     packages: List[ManifestInfo] = []
 
@@ -61,25 +58,19 @@ class ManifestIndex(BaseModel):
     def __hash__(self):
         return hash((self.minimumVersion, frozenset(self.packages)))
 
-    def __eq__(self, other):
-        return hash(self) == hash(other)
 
-
-class SimplePatternList(BaseModel):
+class SimplePatternList(HashEqModel):
     include: List[str] = []
     exclude: List[str] = []
 
     def __hash__(self):
         return hash(frozenset(self.include + self.exclude))
 
-    def __eq__(self, other):
-        return hash(self) == hash(other)
-
     class Config:
         allow_mutation = False
 
 
-class FeatureProperties(BaseModel):
+class FeatureProperties(HashEqModel):
     name: str
     description: str
     recommendation: str = "starred"
@@ -88,9 +79,6 @@ class FeatureProperties(BaseModel):
     def __hash__(self):
         return hash((self.name, self.description, self.recommendation, self.selected))
 
-    def __eq__(self, other):
-        return hash(self) == hash(other)
-
     class Config:
         allow_mutation = False
 
@@ -98,46 +86,20 @@ class FeatureProperties(BaseModel):
 class Feature(BaseModel):
     properties: FeatureProperties
     files: SimplePatternList
+    userFiles: Optional[SimplePatternList] = SimplePatternList()
 
     class Config:
         allow_mutation = False
 
 
-class Task(BaseModel):
+class FileTask(HashEqModel):
     type: str = "file"
     hash: str
     location: str
     to: str
     size: int
-
-    bundle: Optional[str] = None
-
-    def __hash__(self):
-        return hash((
-            self.type,
-            self.hash,
-            self.location,
-            self.to,
-            self.size,
-        ))
-
-    def __eq__(self, other):
-        return hash(self) == hash(other)
-
-    class Config:
-        allow_mutation = False
-        ignore_extra = False
-
-
-class UserFileTask(BaseModel):
-    type: str = "file"
-    hash: str
-    location: str
-    to: str
-    size: int
-
-    userFile: bool = True
-
+    userFile: bool
+    when: Optional[Dict[Any, Any]] = None
     bundle: Optional[str] = None
 
     def __hash__(self):
@@ -148,40 +110,9 @@ class UserFileTask(BaseModel):
             self.to,
             self.size,
             self.userFile,
+            self.when and (self.when['if'], frozenset(self.when['features']))
         ))
 
-    def __eq__(self, other):
-        return hash(self) == hash(other)
-
-    class Config:
-        allow_mutation = False
-        ignore_extra = False
-
-
-class FeatureTask(BaseModel):
-    type: str = "file"
-    hash: str
-    location: str
-    to: str
-    size: int
-
-    when: Dict[Any, Any]
-    bundle: Optional[str] = None
-
-    def __hash__(self):
-        return hash((
-            self.type,
-            self.hash,
-            self.location,
-            self.to,
-            self.size,
-            self.when['if'],
-            frozenset(self.when['features']),
-        ))
-
-    def __eq__(self, other):
-        return hash(self) == hash(other)
-
     class Config:
         allow_mutation = False
         ignore_extra = False
@@ -190,7 +121,6 @@ class FeatureTask(BaseModel):
 class RichManifest(BaseManifest):
     gameVersion: str
     launch: Dict[str, List[str]] = {"flags": []}
-    minimumVersion: int = 2
     librariesLocation: Optional[str] = None
     objectsLocation: str
     userFiles: SimplePatternList = SimplePatternList()
@@ -200,17 +130,23 @@ class RichManifest(BaseManifest):
 
 
 class SourceManifest(RichManifest):
+    minimumVersion: int = 2
     features: List[Feature] = []
     bundles: List[str] = []
 
+    @property
+    def combined_user_files(self) -> Iterable[str]:
+        return chain(self.userFiles.include, *(f.userFiles.include for f in self.features))
+
     class Config:
         allow_mutation = False
         ignore_extra = False
 
 
 class TargetManifest(RichManifest):
+    minimumVersion: int = 3
     features: List[FeatureProperties] = []
-    tasks: List[Union[Task, UserFileTask, FeatureTask]] = []
+    tasks: List[FileTask] = []
 
     def __hash__(self):
         return hash((
@@ -224,9 +160,6 @@ class TargetManifest(RichManifest):
             frozenset(self.tasks),
         ))
 
-    def __eq__(self, other):
-        return hash(self) == hash(other)
-
     class Config:
         allow_mutation = False
         ignore_extra = False
@@ -266,6 +199,3 @@ class PatternTestable(Container):
 
     def __len__(self) -> int:
         return len(self.__patterns)
-
-
-TargetTask = Union[Task, FeatureTask, UserFileTask]
